import pytest

import anyio
from anyioutils import Queue

pytestmark = pytest.mark.anyio


def test_queue_nowait():
    queue = Queue(1)
    assert queue.maxsize == 1
    assert queue.qsize() == 0
    assert queue.empty()
    assert not queue.full()
    queue.put_nowait("foo")
    assert not queue.empty()
    assert queue.full()
    assert queue.qsize() == 1

    with pytest.raises(anyio.WouldBlock):
        queue.put_nowait("bar")

    assert queue.qsize() == 1
    assert queue.get_nowait() == "foo"
    assert queue.qsize() == 0
    assert queue.empty()
    assert not queue.full()


async def test_queue():
    queue = Queue()
    assert queue.maxsize == 0
    assert queue.qsize() == 0
    assert queue.empty()
    assert not queue.full()
    await queue.put("foo")
    assert not queue.empty()
    assert not queue.full()
    assert queue.qsize() == 1
    await queue.put("bar")
    assert queue.qsize() == 2
    assert await queue.get() == "foo"
    assert queue.qsize() == 1
    assert not queue.empty()
    assert not queue.full()
    assert await queue.get() == "bar"
    assert queue.qsize() == 0
    assert queue.empty()
    assert not queue.full()



import asyncio
import queue

import pytest
import trio
from anyioutils import start_guest_run


# copied from: https://trio.readthedocs.io/en/stable/reference-lowlevel.html#implementing-guest-mode-for-your-favorite-event-loop
def test_host_asyncio_guest_trio():
    results = []

    async def trio_main():
        for _ in range(5):
            results.append("Hello from trio!")
            await trio.sleep(0.1)
        return "trio done!"
    
    async def asyncio_main():
        asyncio_loop = asyncio.get_running_loop()
    
        def run_sync_soon_threadsafe(fn):
            asyncio_loop.call_soon_threadsafe(fn)
    
        def done_callback(trio_main_outcome):
            results.append(f"Trio program ended with: {trio_main_outcome}")
    
        start_guest_run(
            trio_main,
            run_sync_soon_threadsafe=run_sync_soon_threadsafe,
            done_callback=done_callback,
            backend="trio",
        )
    
        await asyncio.sleep(1)
    
    asyncio.run(asyncio_main())

    assert results == ["Hello from trio!"] * 5 + ["Trio program ended with: Value('trio done!')"]


# copied from: https://github.com/oremanj/aioguest/blob/b9480a01600d44f8d35e78c7f6689f187331df64/tests/test_basic.py#L52
def test_host_trivial_guest_asyncio():
    results = []
    todo = queue.Queue()

    async def asyncio_main():
        for _ in range(5):
            results.append("Hello from asyncio!")
            await asyncio.sleep(0.1)
        return "asyncio done!"

    def run_sync_soon_threadsafe(fn):
        todo.put(("run", fn))

    def done_callback(result):
        todo.put(("done", result))

    start_guest_run(
        asyncio_main,
        run_sync_soon_threadsafe=run_sync_soon_threadsafe,
        done_callback=done_callback,
        backend="asyncio",
    )

    while True:
        op, obj = todo.get()
        if op == "run":
            obj()  # pragma: nocover
        elif op == "done":
            results.append(f"Asyncio program ended with: {obj}")
            break

    assert results == ["Hello from asyncio!"] * 5 + ["Asyncio program ended with: Value('asyncio done!')"]


def test_wrong_backend():
    async def main(): pass

    def run_sync_soon_threadsafe(fn): pass

    def done_callback(result): pass
    
    with pytest.raises(RuntimeError) as excinfo:
        start_guest_run(
            main,
            run_sync_soon_threadsafe=run_sync_soon_threadsafe,
            done_callback=done_callback,
            backend="foo",
        )

    assert str(excinfo.value) == 'Backend not supported: "foo"'


def test_not_coroutine_function():
    def run_sync_soon_threadsafe(fn): pass

    def done_callback(result): pass
    
    with pytest.raises(RuntimeError) as excinfo:
        start_guest_run(
            None,
            run_sync_soon_threadsafe=run_sync_soon_threadsafe,
            done_callback=done_callback,
            backend="asyncio",
        )

    assert str(excinfo.value) == "Expected a coroutine function, got: None"



from __future__ import annotations

from sys import version_info
from collections.abc import Awaitable, Coroutine
from contextvars import ContextVar
from typing import Any, Callable, Generic, TypeVar

from anyio import Event, create_task_group
from anyio.abc import TaskGroup

from ._exceptions import CancelledError, InvalidStateError
from ._queue import Queue

if version_info < (3, 11):  # pragma: no cover
    from exceptiongroup import BaseExceptionGroup  # type: ignore[import-not-found]


T = TypeVar("T")
_task_group: ContextVar[TaskGroup] = ContextVar("_task_group")


class Task(Generic[T]):
    _done_callbacks: list[Callable[[Task], None]]
    _exception: BaseException | None

    def __init__(self, coro: Coroutine[Any, Any, T]) -> None:
        self._coro = coro
        self._has_result = False
        self._has_exception = False
        self._cancelled_event = Event()
        self._raise_cancelled_error = True
        self._done_callbacks = []
        self._done_event = Event()
        self._exception = None
        self._waiting = False
        self._started_value = Queue[Any]()

    def _call_callbacks(self) -> None:
        exceptions = []
        for callback in self._done_callbacks:
            try:
                callback(self)
            except BaseException as exc:
                exceptions.append(exc)
        if not exceptions:
            return
        if len(exceptions) == 1:
            raise exceptions[0]
        raise BaseExceptionGroup("Error while calling callbacks", exceptions)

    async def _wait_result(self, task_group: TaskGroup) -> None:
        try:
            self._result = await self._coro
            self._has_result = True
        except BaseException as exc:
            self._exception = exc
            self._has_exception = True
        self._done_event.set()
        task_group.cancel_scope.cancel()
        self._call_callbacks()

    async def _wait_cancelled(self, task_group: TaskGroup) -> None:
        await self._cancelled_event.wait()
        task_group.cancel_scope.cancel()

    def cancel(self, raise_exception: bool = False):
        self._done_event.set()
        self._cancelled_event.set()
        self._raise_cancelled_error = raise_exception
        self._call_callbacks()

    def cancelled(self) -> bool:
        return self._cancelled_event.is_set()

    async def wait(self) -> T | None:
        if self._waiting:
            await self._done_event.wait()
        self._waiting = True
        if self._has_result:
            return self._result
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError
            return None
        if self._has_exception:
            assert self._exception is not None
            raise self._exception

        async with create_task_group() as tg:
            tg.start_soon(self._wait_result, tg)
            tg.start_soon(self._wait_cancelled, tg)

        if self._has_result:
            return self._result
        if self._cancelled_event.is_set():
            if self._raise_cancelled_error:
                raise CancelledError
            return None
        if self._has_exception:
            assert self._exception is not None
            raise self._exception

        return None  # pragma: nocover

    def done(self) -> bool:
        return self._done_event.is_set()

    def result(self) -> T:
        if self._cancelled_event.is_set():
            raise CancelledError
        if self._has_result:
            return self._result
        if self._has_exception:
            assert self._exception is not None
            raise self._exception
        raise InvalidStateError

    def exception(self) -> BaseException | None:
        if not self._done_event.is_set():
            raise InvalidStateError
        if self._cancelled_event.is_set():
            raise CancelledError
        return self._exception

    def add_done_callback(self, callback: Callable[[Task], None]) -> None:
        self._done_callbacks.append(callback)
        if self._done_event.is_set():
            callback(self)

    def remove_done_callback(self, callback: Callable[[Task], None]) -> int:
        count = self._done_callbacks.count(callback)
        for _ in range(count):
            self._done_callbacks.remove(callback)
        return count

    async def wait_started(self) -> Any:
        return await self._started_value.get()


def create_task(coro: Coroutine[Any, Any, T], task_group: TaskGroup | None = None, *, name: str | None = None) -> Task[T]:
    task = Task[T](coro)
    if task_group is None:
        task_group = _task_group.get()
    task_group.start_soon(task.wait, name=name)
    return task


async def start_task(async_fn: Callable[..., Awaitable[Any]], task_group: TaskGroup | None = None, *, name: str | None = None) -> Task[None]:
    async_function_wrapper = AsyncFunctionWrapper(async_fn)
    task = Task[None](async_function_wrapper.get_coro())
    async_function_wrapper.set_task(task)
    if task_group is None:
        task_group = _task_group.get()
    task_group.start_soon(task.wait, name=name)
    return task


class AsyncFunctionWrapper:
    def __init__(self, async_fn: Callable[..., Awaitable[T]]) -> None:
        self._async_fn = async_fn

    def set_task(self, task: Task) -> None:
        self._task = task

    async def get_coro(self) -> None:
        async with create_task_group() as tg:
            started_value = await tg.start(self._async_fn)
            await self._task._started_value.put(started_value)



import pytest

from anyio import move_on_after
from anyioutils import Event

pytestmark = pytest.mark.anyio


async def test_event():
    event = Event()
    assert not event.is_set()
    event.set()
    assert event.is_set()
    await event.wait()
    event.clear()
    with move_on_after(0.1):
        await event.wait()
    assert not event.is_set()
